'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _get = function get(object, property, receiver) { if (object === null) object = Function.prototype; var desc = Object.getOwnPropertyDescriptor(object, property); if (desc === undefined) { var parent = Object.getPrototypeOf(object); if (parent === null) { return undefined; } else { return get(parent, property, receiver); } } else if ("value" in desc) { return desc.value; } else { var getter = desc.get; if (getter === undefined) { return undefined; } return getter.call(receiver); } };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _events = require('./events');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var P = Promise;

/**
 * Component is a configurable object that emits events.
 */

var Component = function (_EventEmitter) {
  _inherits(Component, _EventEmitter);

  _createClass(Component, [{
    key: 'defaults',


    /**
     * @param void
     * @return {Object}
     */
    value: function defaults() {
      return {};
    }

    /**
     * @constructor
     * @param {Object} options - instance configuration
     */

  }]);

  function Component() {
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Component);

    var _this = _possibleConstructorReturn(this, (Component.__proto__ || Object.getPrototypeOf(Component)).call(this));

    _this.options = _this.defaults();
    _this.configure(options);
    return _this;
  }

  /**
   * @param {String} key
   * @param {Object} opts
   * @param mixed opts.fallback
   * @param {Boolean} opts.pull
   * @return mixed
   */


  _createClass(Component, [{
    key: 'option',
    value: function option(key) {
      var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var _fallback$pull$opts = _extends({ fallback: undefined, pull: false }, opts),
          fallback = _fallback$pull$opts.fallback,
          pull = _fallback$pull$opts.pull;

      var options = this.options;

      var value = fallback;

      if (!options.hasOwnProperty(key) || options[key] === undefined) {
        if (fallback === undefined) {
          throw new Error('option "' + key + '" is undefined');
        }
      } else {
        value = options[key];
      }

      if (pull) {
        delete options[key];
      }

      return value;
    }

    /**
     * @param {Object} options
     * @return void
     */

  }, {
    key: 'configure',
    value: function configure() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      this.options = _extends({}, this.options, options);
    }

    /**
     * Return a Promise that wraps an event listener. Promise outcomes:
     *  - resolve with event arguments if the event was raised 
     *  - reject if the operation times out
     *
     * @async Promise
     * @param {String} event
     * @param {Number} ttl - optional time-to-live in millis; omit to wait indefinitely
     * @return {Array} - the arguments emitted with the event
     */

  }, {
    key: 'await',
    value: function await(event) {
      var _this2 = this;

      var ttl = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

      return new P(function (resolve, reject) {
        var timeout = void 0;

        var listener = function listener() {
          for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
          }

          clearTimeout(timeout);
          resolve(args);
        };

        var expire = function expire() {
          _this2.off(event, listener);
          reject(new Error('await() operation timed out'));
        };

        if (ttl !== null) {
          timeout = setTimeout(expire, ttl);
        }

        _this2.once(event, listener);
      });
    }
  }]);

  return Component;
}(_events.EventEmitter);

/**
 * Container is a key-value store that has:
 *  - an asynchronous set() method
 *  - a synchronous get() method
 */


var Container = function (_Component) {
  _inherits(Container, _Component);

  _createClass(Container, [{
    key: 'defaults',


    /**
     * @inheritdoc
     */
    value: function defaults() {
      return _extends({}, _get(Container.prototype.__proto__ || Object.getPrototypeOf(Container.prototype), 'defaults', this).call(this), {
        bindings: new Map()
      });
    }

    /**
     * @inheritdoc
     */

  }]);

  function Container() {
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Container);

    var _this3 = _possibleConstructorReturn(this, (Container.__proto__ || Object.getPrototypeOf(Container)).call(this, options));

    _this3.bindings = _this3.option('bindings', { pull: true });
    return _this3;
  }

  /**
   * Bind a value to the container on the supplied key. Value may be any of the
   * following:
   *  - any primitive data type
   *  - a promise
   *  - a function that returns anything, including a promise
   *
   * A promised value will bind to the container when the promise settles.
   *
   * @async Promise
   * @param {String} key
   * @param mixed value
   * @param {Boolean} force
   * @return void
   */


  _createClass(Container, [{
    key: 'set',
    value: function set(key, value) {
      var force = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      var bindings = this.bindings;


      if (bindings.has(key) && !force) {
        throw new Error('bind conflict on key "' + key + '"');
      }

      var bind = function bind() {
        if (!arguments.length) {
          throw new Error('expected an argument');
        }
        bindings.set(key, arguments.length <= 0 ? undefined : arguments[0]);
        return P.resolve();
      };

      if (value && typeof value === 'function') {
        value = value();
      }

      if (value && typeof value.then === 'function') {
        return value.then(bind);
      }

      return bind(value);
    }

    /**
     * @param {String} key
     * @para
     *
     */

  }, {
    key: 'get',
    value: function get(key) {
      var bindings = this.bindings;


      if (!bindings.has(key)) {
        throw new Error('binding not found on key "' + key + '"');
      }

      return bindings.get(key);
    }
  }]);

  return Container;
}(Component);

exports.default = {
  Component: Component,
  Container: Container
};
module.exports = exports['default'];