'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var NotImplemented = new Error('not implemented');

/**
 * Works like PHP's empty() function
 * @param mixed value
 * @return {Boolean}
 */
var empty = function empty(value) {
  if (!value) {
    return value === undefined || !isNaN(value);
  }

  if (typeof value === 'number') {
    return value === 0;
  }

  if (typeof value.length === 'number') {
    return !value.length;
  }

  if (typeof value.size === 'number') {
    return !value.size;
  }

  if ((typeof value === 'undefined' ? 'undefined' : _typeof(value)) === 'object') {
    for (var k in value) {
      if (value.hasOwnProperty(k)) {
        return false;
      }
    }

    return true;
  }

  return false;
};

/**
 * Prototype chaining function
 * @param {Function} subclass
 * @param {Function} superclass
 * @param {Object} augments
 * @return {Function} the subclass
 */
var extend = function extend(subclass, superclass, augments) {
  var Surrogate = function Surrogate() {};

  Surrogate.prototype = superclass.prototype;
  subclass.prototype = new Surrogate();
  subclass.prototype.constructor = subclass;

  assign(subclass.prototype, augments || {});

  return subclass;
};

/**
 * Works like Object.assign()
 */
var assign = function assign() {
  for (var _len = arguments.length, donors = Array(_len), _key = 0; _key < _len; _key++) {
    donors[_key] = arguments[_key];
  }

  var recipient = donors.shift();

  var augment = function augment(recipient, donor) {
    for (var k in donor) {
      if (donor.hasOwnProperty(k)) {
        recipient[k] = donor[k];
      }
    }
    return recipient;
  };

  return donors.reduce(augment, recipient);
};

/**
 * Return a concrete Iterator instance
 * @param {Array|Object|Map|Set} collection
 * @return {Iterator}
 */
var iterator = function iterator(collection) {
  if (collection instanceof Iterator) {
    return collection;
  }
  if (Array.isArray(collection)) {
    return new ArrayIterator(collection);
  }
  if (collection instanceof Map) {
    return new MapIterator(collection);
  }
  if (collection instanceof Set) {
    return new SetIterator(collection);
  }
  if (collection && (typeof collection === 'undefined' ? 'undefined' : _typeof(collection)) === 'object') {
    return new ObjectIterator(collection);
  }

  throw new Error('expected iterable argument');
};

/**
 * Return an array of keys in the collection
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
var keys = function keys(collection) {
  return iterator(collection).keys();
};

/**
 * Return an array of values in the collection
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
var values = function values(collection) {
  return iterator(collection).values();
};

/**
 * Return the size of the collection
 * @param {Array|Object|Map|Set} collection
 * @return Number
 */
var size = function size(collection) {
  return iterator(collection).size();
};

/**
 * Return a representation of the collection as a list of key-value tuples:
 *
 *   [['key1', 'value1'], ['key2', 'value2'], ..., ['keyN', 'valueN']]
 *
 * @param {Array|Object|Map|Set} collection
 * @return {Array}
 */
var entries = function entries(collection) {
  return iterator(collection).entries();
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - reducer function
 * @param mixed initial
 * @return mixed
 */
var reduce = function reduce() {
  return reduce.until.apply(reduce, arguments);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - reducer function
 * @param mixed initial - the initial carry
 * @param {Function} breakpoint - reduce until this function returns true
 * @return mixed
 */
reduce.until = function (collection, fn, initial) {
  var breakpoint = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : function () {
    return false;
  };

  var pairs = entries(collection);
  var out = initial;

  while (pairs.length) {
    var _pairs$shift = pairs.shift(),
        _pairs$shift2 = _slicedToArray(_pairs$shift, 2),
        k = _pairs$shift2[0],
        v = _pairs$shift2[1];

    if (breakpoint(out, v, k, collection)) {
      break;
    }

    out = fn(out, v, k, collection);
  }

  return out;
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - map function
 * @return {Array}
 */
var map = function map() {
  return map.until.apply(map, arguments);
};

/**
 * @param {Array}Object|Map|Set} collection
 * @param {Function} fn - map function
 * @param {Function} breakpoint - reduce until this function returns true
 * @return {Array}
 */
map.until = function (collection, fn) {
  var breakpoint = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : function () {
    return false;
  };

  var reducer = function reducer(out) {
    for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
      args[_key2 - 1] = arguments[_key2];
    }

    out.push(fn.apply(undefined, args));
    return out;
  };

  var test = function test(out) {
    for (var _len3 = arguments.length, args = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
      args[_key3 - 1] = arguments[_key3];
    }

    return breakpoint.apply(undefined, args);
  };
  return reduce.until(collection, reducer, [], test);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - iteration callback
 * @return void
 */
var each = function each() {
  return each.until.apply(each, arguments);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - iteration callback
 * @param {Function} breakpoint - iterate until this function returns true
 * @return void
 */
each.until = function (collection, fn) {
  var breakpoint = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : function () {
    return false;
  };

  map.until(collection, fn, breakpoint);
};

/**
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - filter function
 * @return {Array}
 */
var filter = function filter(collection, fn) {
  var pairs = entries(collection);
  var out = [];

  while (pairs.length) {
    var _pairs$shift3 = pairs.shift(),
        _pairs$shift4 = _slicedToArray(_pairs$shift3, 2),
        k = _pairs$shift4[0],
        v = _pairs$shift4[1];

    fn(v, k, collection) && out.push(v);
  }

  return out;
};

/**
 * Return the first value in the collection that passes a truth test
 * Return the very first value in the collection if a truth test is not supplied
 * @param {Array|Object|Map|Set} collection
 * @param {Function} fn - optional truth test
 * @return mixed
 */
var first = function first(collection) {
  var fn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (v) {
    return true;
  };

  var pairs = entries(collection);

  while (pairs.length) {
    var _pairs$shift5 = pairs.shift(),
        _pairs$shift6 = _slicedToArray(_pairs$shift5, 2),
        k = _pairs$shift6[0],
        v = _pairs$shift6[1];

    if (fn(v, k, collection)) {
      return v;
    }
  }

  return null;
};

/**
 * @param {Array|Object|Map|Set} collection
 * @return {Array|Object|Map|Set)
 */
var reverse = function reverse(collection) {
  return iterator(collection).reverse();
};

exports.default = {
  empty: empty,
  extend: extend,
  assign: assign,
  keys: keys,
  values: values,
  size: size,
  entries: entries,
  reduce: reduce,
  each: each,
  map: map,
  filter: filter,
  first: first,
  reverse: reverse
};

////////////////////////////////////////////////////////////////////////////////

var Iterator = function () {
  function Iterator() {
    _classCallCheck(this, Iterator);
  }

  _createClass(Iterator, [{
    key: 'keys',


    /**
     * Return an Array
     * @param void
     * @return {Array}
     */
    value: function keys() {
      throw NotImplemented;
    }

    /**
     * Return an Array
     * @param void
     * @return {Array}
     */

  }, {
    key: 'values',
    value: function values() {
      throw NotImplemented;
    }

    /**
     * Return an Array of two-element tuples: [key, value]
     * @param void
     * @return {Array}
     */

  }, {
    key: 'entries',
    value: function entries() {
      throw NotImplemented;
    }

    /**
     * @param void
     * @return {Number}
     */

  }, {
    key: 'size',
    value: function size() {
      throw NotImplemented;
    }

    /**
     * Return a reversed representation of the collection
     * @param void
     * @return mixed
     */

  }, {
    key: 'reverse',
    value: function reverse() {
      throw NotImplemented;
    }
  }]);

  return Iterator;
}();

var ArrayIterator = function (_Iterator) {
  _inherits(ArrayIterator, _Iterator);

  /**
   * @param {Array} array
   */
  function ArrayIterator(array) {
    _classCallCheck(this, ArrayIterator);

    var _this = _possibleConstructorReturn(this, (ArrayIterator.__proto__ || Object.getPrototypeOf(ArrayIterator)).call(this));

    _this.array = array;
    return _this;
  }

  /**
   * @inheritdoc
   */


  _createClass(ArrayIterator, [{
    key: 'keys',
    value: function keys() {
      var out = [];
      var size = this.size();
      var i = 0;

      while (i < size) {
        out.push(i);
        i += 1;
      }

      return out;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'values',
    value: function values() {
      var array = this.array;

      var out = [];
      var size = this.size();
      var i = 0;

      while (i < size) {
        out.push(array[i]);
        i += 1;
      }

      return out;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'entries',
    value: function entries() {
      var array = this.array;

      var out = [];
      var size = this.size();
      var i = 0;

      while (i < size) {
        out.push([i, array[i]]);
        i += 1;
      }

      return out;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'size',
    value: function size() {
      return this.array.length;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'reverse',
    value: function reverse() {
      return this.array.reverse();
    }
  }]);

  return ArrayIterator;
}(Iterator);

var ObjectIterator = function (_Iterator2) {
  _inherits(ObjectIterator, _Iterator2);

  /**
   * @param {Object} object
   */
  function ObjectIterator(object) {
    _classCallCheck(this, ObjectIterator);

    var _this2 = _possibleConstructorReturn(this, (ObjectIterator.__proto__ || Object.getPrototypeOf(ObjectIterator)).call(this));

    _this2.object = object;
    return _this2;
  }

  /**
   * @inheritdoc
   */


  _createClass(ObjectIterator, [{
    key: 'keys',
    value: function keys() {
      var object = this.object;

      var out = [];

      for (var k in object) {
        object.hasOwnProperty(k) && out.push(k);
      }

      return out;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'values',
    value: function values() {
      var object = this.object;

      var out = [];

      for (var k in object) {
        object.hasOwnProperty(k) && out.push(object[k]);
      }

      return out;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'entries',
    value: function entries() {
      var object = this.object;

      var out = [];

      for (var k in object) {
        object.hasOwnProperty(k) && out.push([k, object[k]]);
      }

      return out;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'size',
    value: function size() {
      return this.keys().length;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'reverse',
    value: function reverse() {
      return this.entries().reverse().reduce(function (out, v, k) {
        return out[k] = v;
      }, {});
    }
  }]);

  return ObjectIterator;
}(Iterator);

var ES6CollectionIterator = function (_Iterator3) {
  _inherits(ES6CollectionIterator, _Iterator3);

  /**
   * @param {Map|Set} collection
   */
  function ES6CollectionIterator(collection) {
    _classCallCheck(this, ES6CollectionIterator);

    var _this3 = _possibleConstructorReturn(this, (ES6CollectionIterator.__proto__ || Object.getPrototypeOf(ES6CollectionIterator)).call(this));

    _this3.collection = collection;
    return _this3;
  }

  /**
   * @inheritdoc
   */


  _createClass(ES6CollectionIterator, [{
    key: 'keys',
    value: function keys() {
      var keys = [];
      this.collection.forEach(function (v, k) {
        return keys.push(k);
      });
      return keys;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'values',
    value: function values() {
      var values = [];
      this.collection.forEach(function (v, k) {
        return values.push(v);
      });
      return values;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'entries',
    value: function entries() {
      var entries = [];
      this.collection.forEach(function (v, k) {
        return entries.push([k, v]);
      });
      return entries;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'size',
    value: function size() {
      return this.collection.size;
    }

    /**
     * @inheritdoc
     */

  }, {
    key: 'reverse',
    value: function reverse() {
      var constructor = this.collection.constructor;
      return new constructor(this.entries().reverse());
    }
  }]);

  return ES6CollectionIterator;
}(Iterator);

var MapIterator = function (_ES6CollectionIterato) {
  _inherits(MapIterator, _ES6CollectionIterato);

  function MapIterator() {
    _classCallCheck(this, MapIterator);

    return _possibleConstructorReturn(this, (MapIterator.__proto__ || Object.getPrototypeOf(MapIterator)).apply(this, arguments));
  }

  return MapIterator;
}(ES6CollectionIterator);

var SetIterator = function (_ES6CollectionIterato2) {
  _inherits(SetIterator, _ES6CollectionIterato2);

  function SetIterator() {
    _classCallCheck(this, SetIterator);

    return _possibleConstructorReturn(this, (SetIterator.__proto__ || Object.getPrototypeOf(SetIterator)).apply(this, arguments));
  }

  return SetIterator;
}(ES6CollectionIterator);

module.exports = exports['default'];